/**
 * gclc-socket:net.bigeon.gclc.socket.DConnexionManager.java
 * Created on: Nov 18, 2017
 */
package net.bigeon.gclc.socket;

/*-
 * #%L
 * GCLC Socket
 * %%
 * Copyright (C) 2016 - 2018 Bigeon
 * %%
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * #L%
 */
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/** Default implementation of the {@link ConnexionManager}.
 *
 * @author Emmanuel Bigeon
 * @param <T> the connected objects */
public final class DConnexionManager<T> implements ConnexionManager<T> {

    /** Class logger. */
    private static final Logger        LOGGER      = Logger
            .getLogger(DConnexionManager.class.getName());
    /** The connected objects. */
    private final Map<String, T>       connecteds  = new HashMap<>();
    /** The locks for the connexions. */
    private final Map<String, Object>  locks       = new HashMap<>();
    /** The counter for the disconnexion locks. */
    private final Map<String, Integer> counters    = new HashMap<>();
    /** The lock for modification of {@link #counters}. */
    private final Object               counterLock = new Object();
    /** The count of connexions. */
    private int                        count;

    /** Default.constructor. */
    public DConnexionManager() {
        //
    }

    /* (non-Javadoc)
     * @see
     * net.bigeon.gclc.socket.ConnexionManager#addConnexion(java.lang.Object) */
    @Override
    public String addConnexion(final T handle) {
        final String newID = newID();
        connecteds.put(newID, handle);
        locks.put(newID, new Object());
        counters.put(newID, Integer.valueOf(0));
        return newID;
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.socket.ConnexionManager#disconnect(java.lang.String) */
    @Override
    public T disconnect(final String id) {
        if (!connecteds.containsKey(id)) {
            return null;
        }
        final T disc = connecteds.remove(id);
        final Object lock = locks.get(id);
        synchronized (lock) {
            lock.notifyAll();
        }
        synchronized (counterLock) {
            while (counters.get(id).intValue() > 0) {
                try {
                    counterLock.wait();
                } catch (final InterruptedException e) {
                    LOGGER.log(Level.FINE, "Interruption of thread", e); //$NON-NLS-1$
                    Thread.currentThread().interrupt();
                }
            }
        }
        return disc;
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.socket.ConnexionManager#get(java.lang.String) */
    @Override
    public T get(final String id) {
        return connecteds.get(id);
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.socket.ConnexionManager#getConnected() */
    @Override
    public Collection<String> getConnected() {
        return connecteds.keySet();
    }

    /* (non-Javadoc)
     * @see net.bigeon.gclc.socket.ConnexionManager#isConnected(java.lang.String) */
    @Override
    public boolean isConnected(final String id) {
        return connecteds.containsKey(id);
    }

    /* (non-Javadoc)
     * @see
     * net.bigeon.gclc.socket.ConnexionManager#lockDisconnexion(java.lang.String) */
    @Override
    public void lockDisconnexion(final String id) {
        if (!connecteds.containsKey(id)) {
            return;
        }
        synchronized (counterLock) {
            counters.put(id, Integer.valueOf(counters.get(id).intValue() + 1));
        }
    }

    /** Get a new identifier for connexion.
     *
     * @return a new ID */
    private String newID() {
        final int c = count++;
        return "Client " + c; //$NON-NLS-1$
    }

    /* (non-Javadoc)
     * @see
     * net.bigeon.gclc.socket.ConnexionManager#releaseDisconnexionLock(java.lang.
     * String) */
    @Override
    public void releaseDisconnexionLock(final String id) {
        synchronized (counterLock) {
            counters.put(id,
                    Integer.valueOf(Math.max(counters.get(id).intValue() - 1, 0)));
            counterLock.notifyAll();
        }
    }

    /* (non-Javadoc)
     * @see
     * net.bigeon.gclc.socket.ConnexionManager#waitDisconnexion(java.lang.String) */
    @Override
    public void waitDisconnexion(final String id) throws InterruptedException {
        final Object lock = locks.get(id);
        while (connecteds.containsKey(id)) {
            synchronized (lock) {
                lock.wait();
            }
        }
    }
}
